# coding=utf-8
import os
from click.testing import CliRunner
import unittest

from Stackml.Cli import cli
from Stackml_tests.acceptances.Fixtures import clone_fixture


class MainTest(unittest.TestCase):

    def setUp(self):
        pass


    def test_compile_exit_error_code_if_action_fails_not_0(self):
        # Assign
        runner = CliRunner()

        # Acts
        result = runner.invoke(cli, ['compile', '--input', '--backend', 'backend', '--output', 'outputfile'])

        # Assert
        self.assertNotEqual(result.exit_code, 0)

    # def test_compile_stackcore_should_generate_instance_file_when_stackml_architecture_is_valid(self):
    #     # Assign
    #     runner = CliRunner()
    #     with clone_fixture('empty_stackml_architecture') as working_directory:
    #         os.chdir(working_directory)
    #         stackml_architecture = '.'
    #         output_directory = 'output'
    #
    #         # Acts
    #         result = runner.invoke(cli, ['compile', '--input', stackml_architecture, 'stackcore',
    #                                      '--output', output_directory], input='y\n ')
    #
    #         # Assert
    #         ist_output_file = os.path.join(output_directory, 'Instance.ist.xml')
    #
    #         self.assertEqual(result.exit_code, 0, result)
    #         self.assertTrue(os.path.isfile(ist_output_file), f'the Instance file does not exists { ist_output_file }')
    #
    # def test_compile_stackcore_should_generate_initialization_file_when_stackml_architecture_is_valid(self):
    #     # Assign
    #     runner = CliRunner()
    #     with clone_fixture('empty_stackml_architecture') as working_directory:
    #         os.chdir(working_directory)
    #         stackml_architecture = '.'
    #         output_directory = 'output'
    #
    #         # Acts
    #         result = runner.invoke(cli, ['compile', '--input', stackml_architecture, 'stackcore',
    #                                      '--output', output_directory], input='y\n ')
    #
    #         # Assert
    #         ist_output_file = os.path.join(os.getcwd(), output_directory, 'Initialization.ini.xml')
    #
    #         self.assertEqual(result.exit_code, 0, result)
    #         self.assertTrue(os.path.isfile(ist_output_file),
    #                         f'the Initialization file does not exists { ist_output_file }')
    #
    #
    # def test_compile_stackcore_should_remove_output_folder_when_output_is_not_empty_and_user_accept(self):
    #     # Assign
    #     runner = CliRunner()
    #     with clone_fixture('non_empty_output_folder') as working_directory:
    #         os.chdir(working_directory)
    #         stackml_architecture = '.'
    #         output_directory = 'output'
    #
    #         # Acts
    #         result = runner.invoke(cli, ['compile', '--input', stackml_architecture, 'stackcore',
    #                                      '--output', output_directory], input='y\n')
    #
    #         # Assert
    #         output_content = os.path.join(os.getcwd(), output_directory, 'file.txt')
    #
    #         self.assertEqual(result.exit_code, 0, result)
    #         self.assertFalse(os.path.isfile(output_content),
    #                          f'the output directory is not be cleaned { output_content }')
    #
    # def test_compile_stackcore_should_not_remove_output_folder_when_output_is_not_empty_and_user_decline(self):
    #     # Assign
    #     runner = CliRunner()
    #     with clone_fixture('non_empty_output_folder') as working_directory:
    #         os.chdir(working_directory)
    #         stackml_architecture = '.'
    #         output_directory = 'output'
    #
    #         # Acts
    #         result = runner.invoke(cli, ['compile', '--input', stackml_architecture, 'stackcore',
    #                                      '--output', output_directory], input='N\n y\n output1')
    #
    #         # Assert
    #         output_content = os.path.join(os.getcwd(), output_directory, 'file.txt')
    #
    #         self.assertEqual(result.exit_code, 0, result)
    #
    #         self.assertTrue(os.path.isfile(output_content))
    #
    # def test_compile_stackcore_should_create_output_folder_when_output_is_not_found(self):
    #     # Assign
    #     runner = CliRunner()
    #     with clone_fixture('empty_stackml_architecture') as working_directory:
    #         os.chdir(working_directory)
    #         stackml_architecture = '.'
    #         output_directory = 'output'
    #
    #         # Acts
    #         result = runner.invoke(cli, ['compile', '--input', stackml_architecture, 'stackcore',
    #                                      '--output', output_directory], input='y\n ')
    #         # Assert
    #         output_content = os.path.join(os.getcwd(), output_directory)
    #
    #         self.assertEqual(result.exit_code, 0, result)
    #         self.assertTrue(os.path.isdir(output_content))

    def test_compile_should_show_error_when_input_file_does_not_exist(self):
        # Assign
        runner = CliRunner()
        with clone_fixture('empty_stackml_architecture') as working_directory:
            os.chdir(working_directory)
            stackml_architecture = 'StackMl_'
            output_file = 'output/test'

            # Acts
            result = runner.invoke(cli, ['compile', '--input', stackml_architecture, 'diagram',
                                         '--type', '1', '--output', output_file])

            # Assert
            self.assertEqual(result.exit_code, 2, result)
            self.assertEqual(result.output.split(stackml_architecture)[1],
                             f' doesn\'t exist\n', f'the input file {stackml_architecture}'
                                                   f' does not exist and the exception hase been raised ')

    def test_compile_should_show_error_when_backend_archeitecture_does_not_exist(self):
        # Assign
        runner = CliRunner()
        with clone_fixture('empty_stackml_architecture') as working_directory:
            os.chdir(working_directory)
            stackml_architecture = '.'
            output_directory = 'output'
            backend = 'backend'

            # Acts
            result = runner.invoke(cli, ['compile', '--input', stackml_architecture, backend])
            # Assert
            self.assertEqual(result.exit_code, 2, result)
            self.assertIn(f"Error: No such command \'{backend}\'.", result.output)


    def test_compile_should_show_error_when_input_file_is_not_well_formed(self):
        # Assign
        runner = CliRunner()
        with clone_fixture('not_well_formed_stackml_architecture') as working_directory:
            os.chdir(working_directory)
            stackml_architecture = '.'
            output_file = 'output/test'

            # Acts
            result = runner.invoke(cli, ['compile', '--input', stackml_architecture, 'diagram',
                                         '--type', '1', '--output', output_file])

            # Assert
            self.assertEqual(result.exit_code, 1, result)
            self.assertIn('YAML is not valid', result.output)

    def test_compile_should_show_error_when_input_file_does_not_respect_stackml_schema(self):
        # Assign
        runner = CliRunner()
        with clone_fixture('not_valid_stackml_schema') as working_directory:
            os.chdir(working_directory)
            stackml_architecture = '.'
            output_file = 'output/test'

            # Acts
            result = runner.invoke(cli, ['compile', '--input', stackml_architecture, 'diagram',
                                         '--type', '1', '--output', output_file])

            # Assert
            self.assertEqual(result.exit_code, 1, result)
            self.assertIn("Schema validation No", result.output)

    def test_check_should_show_error_when_input_file_does_not_exist(self):
        # Assign
        runner = CliRunner()
        with clone_fixture('empty_stackml_architecture') as working_directory:
            os.chdir(working_directory)
            stackml_architecture = 'StackMl_.xml'

            # Acts
            result = runner.invoke(cli, ['check', '--input', stackml_architecture])

            # Assert
            self.assertEqual(result.exit_code, 2, result)
            self.assertEqual(result.output.split(stackml_architecture)[1], f" doesn\'t exist\n")

    def test_check_should_show_error_when_input_file_is_not_well_formed(self):
        # Assign
        runner = CliRunner()
        with clone_fixture('not_well_formed_stackml_architecture') as working_directory:
            os.chdir(working_directory)
            stackml_architecture = '.'

            # Acts
            result = runner.invoke(cli, ['check', '--input', stackml_architecture])

            # Assert
            self.assertEqual(result.exit_code, 1, result)
            self.assertIn('YAML is not valid', result.output)


    def test_init_should_generate_stackml_file(self):
        # Assign
        runner = CliRunner()
        with clone_fixture('stackml_project') as working_directory:
            os.chdir(working_directory)
            stackml_architecture = 'Test'

            # Acts
            result = runner.invoke(cli, ['init', '--path', stackml_architecture], input=' y\n IoT System\nsys1\nuseCas1\n')

            # Assert
            stackml_file = os.path.join(stackml_architecture, 'main.stkml.yaml')

            self.assertEqual(result.exit_code, 0, result)
            self.assertTrue(os.path.isfile(stackml_file), f'the Stackml file does not exists { stackml_file }')


    def test_compile_diagram_should_generate_image_system_diagram(self):
        # Assign
        runner = CliRunner()
        with clone_fixture('diagram_level_1') as working_directory:
            os.chdir(working_directory)
            stackml_architecture = 'Test'
            output_file = 'img/test'

            # Acts
            result = runner.invoke(cli, ['compile', '--input', stackml_architecture, 'diagram',
                                         '--type', '1', '--output', output_file])

            # Assert
            output_file = f'{output_file}.png'
            self.assertEqual(result.exit_code, 0, result)
            self.assertTrue(os.path.isfile(output_file), f'the image file does not exists {output_file}')




    def test_compile_diagram_should_generate_image_for_topology_diagram(self):
        # Assign
        runner = CliRunner()
        with clone_fixture('diagram_tpology') as working_directory:
            os.chdir(working_directory)
            stackml_architecture = 'Test'
            output_file = 'img/test'

            # Acts
            result = runner.invoke(cli, ['compile', '--input', stackml_architecture, 'diagram',
                                         '--type', '2', '--output', output_file])

            # Assert
            output_file = f'{output_file}.png'
            self.assertEqual(result.exit_code, 0, result)
            self.assertTrue(os.path.isfile(output_file), f'the image file does not exists {output_file}')

    def test_compile_diagram_should_generate_map_for_topology_diagram(self):
        # Assign
        runner = CliRunner()
        with clone_fixture('diagram_tpology') as working_directory:
            os.chdir(working_directory)
            stackml_architecture = 'Test'
            output_file = 'img/test'

            # Acts
            result = runner.invoke(cli, ['compile', '--input', stackml_architecture, 'diagram',
                                         '--type', '3', '--output', output_file])

            # Assert
            output_file = f'{output_file}.html'
            self.assertEqual(result.exit_code, 0, result)
            self.assertTrue(os.path.isfile(output_file), f'the html file does not exists {output_file}')


    def test_compile_drawio_should_generate_drawio_file_for_architecture_diagram_level_1(self):
        # Assign
        runner = CliRunner()
        with clone_fixture('diagram_level_1') as working_directory:
            os.chdir(working_directory)
            stackml_architecture = 'Test'
            output_file = 'drawio/test'

            # Acts
            result = runner.invoke(cli, ['compile', '--input', stackml_architecture, 'drawio',
                                         '--level', '1', '--output', output_file])

            # Assert
            output_file = f'{output_file}.drawio'
            self.assertEqual(result.exit_code, 0, result)
            self.assertTrue(os.path.isfile(output_file), f'the drawio file does not exists {output_file}')

    def test_compile_drawio_should_generate_drawio_file_for_architecture_diagram_level_2(self):
        # Assign
        runner = CliRunner()
        with clone_fixture('diagram_level_2') as working_directory:
            os.chdir(working_directory)
            stackml_architecture = 'Test'
            output_file = 'drawio/test'

            # Acts
            result = runner.invoke(cli, ['compile', '--input', stackml_architecture, 'drawio',
                                         '--level', '2', '--output', output_file])

            # Assert
            output_file = f'{output_file}.drawio'
            self.assertEqual(result.exit_code, 0, result)
            self.assertTrue(os.path.isfile(output_file), f'the drawio file does not exists {output_file}')



#Copyright STACKEO - INRIA 2020 .
import os

from Stackml import MODULE_DIR
from Stackml.Diagram.Graphviz.LayerDiagram import LayerDiagram
from Stackml.Diagram.Graphviz.RegionDiagram import RegionDiagram
from Stackml.Diagram.Graphviz.SystemDiagram import SystemDiagram
from Stackml.Diagram.Graphviz.Node import DiagramNode
from Stackml.Diagram.Graphviz.Layer import Diagramlayer
from Stackml.Diagram.Map.NodeMarker import NodeMarker
from Stackml.Diagram.Map.TopologyMap import TopologyMap
from Stackml.Stackml import Stackml


class StackmlDiagram:

    def __init__(self, icons: str):
        self._icon_dir = os.path.join(os.getcwd(), icons)
        self.default_icons = os.path.join(MODULE_DIR, 'templates/diagram/resources/')
        self.diagram_attr = {"pad": "0.2", "splines": "ortho", "nodesep": "0.6",
                             "ranksep": "1", "fontname": "Sans-Serif", "fontsize": "15", "fontcolor": "#2D3436"}

        self.funcs = {1: SystemDiagram, 2: RegionDiagram, 3: TopologyMap}

    def diagram_from_stackml(self, type_: int, stackml: Stackml, output: str) -> str:
        bound = self.funcs[type_](self.default_icons, self._icon_dir, self.diagram_attr)
        return bound.from_stackml(stackml, output)



from unittest import TestCase

from Stackml.Drawio import Node


class TestNode(TestCase):

    def setUp(self) -> None:
        self.node = Node(name='Node', type_='ThingNode', nature='Hardware')

    def test_set_dimension_should_get_the_require_dimension_to_put_all_layers_in_the_node(self):
        # Assign
        self.node.add_layer(type_='Physical', hub=False)
        l = self.node.add_layer(type_='Network', hub=False)

        #Act
        self.node.generate_layers_position()
        self.node.set_dimensions()

        #Assert
        self.assertEqual(self.node.dimension[0], l.dimension[0] + 20)
        self.assertEqual(self.node.dimension[1], l.position[1] + l.dimension[1] + 10)



import unittest

from Stackml.Stackml.LayerElement.BusinesslayerElement import BusinesslayerElement


class TestBusinessLayerElement(unittest.TestCase):

    def setUp(self) -> None:
        pass

    def test_create_business_layer_element(self):
        # Assign
        element = 'Buz'

        # Act
        e = BusinesslayerElement(element, hub='EndPoint')

        # Assert
        self.assertIsInstance(e, BusinesslayerElement)



#Copyright STACKEO - INRIA 2020 .
import os
from typing import List
import copy
import json

import requests
from jinja2 import Environment, Template


import jsonschema
from jsonschema import validate
import yaml
from yaml.scanner import ScannerError

from Stackml import STACKML_URL_SCHEMA, MODULE_DIR
from Stackml.Stackml.LayerElement import Component
from Stackml.Stackml.ModelException import ModelException

from Stackml.Stackml.Region import Region
from Stackml.Stackml.StackmlSyntaxicError import StackmlSyntaxicError
from Stackml.Stackml.StackmlSyntaxicErrorList import StackmlSyntaxicErrorList
from Stackml.Stackml.Tier import Tier
from Stackml.Stackml.Link import Link
from Stackml.Stackml.Node import Node
from Stackml.Stackml.Topology import Topology
from Stackml.Stackml.WarningList import WarningList


class Stackml:

    def __init__(self, folder=None):
        self.stackmlfile = None
        if folder:
            self.stackmlfile = os.path.join(folder, 'main.stkml.yaml')
        self.tiers: List[Tier] = []
        self.links: List[Link] = []
        self.regions: List[Region] = []
        self.topology = Topology()
        self.schema = self.get_stackml_schema()
        self.errors = []
        self.libraries = {}

    def check_file(self, file) -> dict or ScannerError:
        stackml_yaml = yaml.load(file, Loader=yaml.FullLoader)
        file.close()
        return stackml_yaml

    def get_library(self, file):#-> FileNotFoundError or str
        try:
            stackml_file = open(file, 'r')
        except FileNotFoundError:
            file = file.split('/')[-1]
            file = os.path.join(MODULE_DIR, 'templates', 'Stackml', file)
            print(file)
            try:
                stackml_file = open(file, 'r')
            except FileNotFoundError as error:
                raise FileNotFoundError(f'\nThe {file.split("/")[-1].split(".")[0]} library does not exist') from error
        return stackml_file

    def get_stackml_schema(self) -> dict or None:
        try:
            schema = None
            request = requests.get(STACKML_URL_SCHEMA)
        except requests.exceptions.ConnectionError :
            pass
        else:
            if request.status_code == 200:
                schema = json.loads(request.content.decode("utf-8"))
            request.close()
            return schema
    def validate_schema(self, file) -> dict or Exception:
        with open(file, 'r') as file_:
            file_ = yaml.load(file_, Loader=yaml.FullLoader)
            try:
                validate(file_, self.schema)
            except jsonschema.exceptions.ValidationError:
                self.errors.append(StackmlSyntaxicError(self.schema, file_, file))
            return file_


    def check_stackml_project(self) -> str or ScannerError:
        file = self.get_library(self.stackmlfile)
        stackml = self.check_file(file)
        path = os.path.dirname(self.stackmlfile)
        i = 0
        for import_ in (stackml.get('import') or []):
            imports = os.path.join(path, f'{import_}.stkml.yaml')
            imports_ = self.get_library(imports)
            self.libraries[import_] = imports_.name
            self.check_file(imports_)
            i += 1
        file.close()

    def validate_stackml_project(self) -> str or StackmlSyntaxicErrorList or None:
        if self.schema:
            stackml = self.validate_schema(self.stackmlfile)
            for import_ in (stackml.get('import') or []):
                self.validate_schema(self.libraries[import_])
            if len(self.errors) > 0:
                raise StackmlSyntaxicErrorList(self.errors)
            return None
        return 'Stackml schema does not found please verify your connection'

    def add_tier(self, tier_type: str) -> Tier:
        for tier in self.tiers:
            if tier.type == tier_type:
                return tier
        tier = Tier(name=tier_type, type_=tier_type)
        self.tiers.append(tier)
        return tier

    def create_architecture(self):
        with open(self.stackmlfile, 'r') as stackml_file:
            stackml = yaml.load(stackml_file, Loader=yaml.FullLoader)
            topology = stackml['topology']
            self.topology.name = topology.get('name')
            self.topology.usecase = topology.get('useCase')
            self.topology.type = topology.get('type')
            nodes = topology.get('Nodes')
            if nodes:
                self.create_tiers(nodes)
            links = topology.get('Links')
            if links:
                self.create_links(links)
            regions = topology.get('Regions')
            if regions:
                self.create_regions(regions)


    def create_regions(self, regions) -> None:
        for region in regions:
            region_ = Region(name=region.get('name'), type_=region.get('type'))
            for node in region.get('NodePopulation'):
                region_.add_node(node_id=node.get('nodeId'), number=node.get('number'))
            self.regions.append(region_)

    def create_links(self, links) -> None:
        for link in links:
            source = link.get('source')
            sink = link.get('sink')
            self.links.append(Link(source=source, sink=sink,
                                   layer=link.get('layer'), secure=link.get('secure')))
    def get_node_model(self, model_name):
        node = None
        if '.' in model_name:#extrenel model

            model = model_name.split('.')
            file = self.libraries.get(model[0])
            if file:
                with open(file, 'r') as s_file:
                    model_def = self.check_file(s_file)
                node_id = model[1]
            else:
                raise ModelException(msg=f'{model[0]} is not imported')
        else:#internal model
            file = self.stackmlfile
            with open(file, 'r') as s_file:
                model_def = self.check_file(s_file)
            node_id = model_name
        model_def = model_def.get('modeldef')
        if model_def:
            model_def = model_def.get('NodeModels')
            if model_def:
                for node_def in model_def:
                    if node_def.get('name') == node_id:
                        node = node_def
                        break
            else:
                raise ModelException(msg=f'The are no NodeModels on {file}')
        else:
            raise ModelException(msg=f'The are no modeldef for {node_id}')
        if not node:
            raise ModelException(msg=f'The {node_id} model does not exist on {file}')
        return node


    def create_tiers(self, nodes) -> None:
        #warnings = []
        for node in nodes:
            model = node.get('model')
            node_def = self.get_node_model(model)
            node_type = node_def.get('technology')
            tier = node.get('tier')
            #if tier not in node_type:
            #    warnings.append(f"you are using {node_type} in {tier} Tier")
            tier = self.add_tier(tier)
            node_ = tier.add_node(id_=node.get('id'), type_=node_type,
                                  cardinality=node.get('cardinality'))
            node_.set_nature(node_def.get('nature'))
            for layer in node_.element_types:
                elem = node_def.get(layer)
                if elem:
                    hub = elem[0].get('ElementType')
                    if not hub:
                        hub = 'EndPoint'
                    else:
                        elem = elem[1:]
                    layer_elem = node_.add_layer_element(layer, type_=layer, hub=hub)
                    layer_elem.set_components(elem)
        self.sort_tiers()
        #return warnings

    def generate_files(self, environment: Environment) -> Template:
        template = environment.get_template('StackmlIn/StackMl.yaml.j2')
        return template.render(Stackml=self)

    def get_node(self, id_: str) -> Node:
        node = None
        for tier in self.tiers:
            for node_ in tier.nodes:
                if node_.id_ == id_:
                    node = node_
                    break
        return node


    def get_node_tier(self, id_: str) -> Tier or None:
        tier = None
        for tier_ in self.tiers:
            for node in tier_.nodes:
                if node.id_ == id_:
                    tier = tier_
                    break
        return tier

    def get_tier(self, tier_name: str) -> Tier or None:
        tier = None
        for tier_ in self.tiers:
            if tier_.name == tier_name:
                tier = tier_
                break
        return tier

    def get_component(self, layer, cmp_id) -> Component or None:
        comp = None
        for tier in self.tiers:
            for node in tier.nodes:
                cmps = node.get_components(layer)
                for component in (cmps or []):
                    if component.id_ == cmp_id:
                        comp = component
                        break
                else:
                    continue
                break
        return comp

    def creat_referenced_components(self) -> None:
        for i_t in range(len(self.tiers)):
            tier = self.tiers[i_t]
            for i_n in range(len(tier.nodes)):
                node = tier.nodes[i_n]
                for i_layer in range(len(node.layers_element)):
                    layer = node.layers_element[i_layer]
                    for i_comp in range(len(layer.components)):
                        comp = layer.components[i_comp]
                        id_ = comp.id_ref
                        if id_:

                            comp = self.get_component(layer.type, id_)
                            id_ = comp.id_
                            comp = copy.deepcopy(comp)
                            layer.components[i_comp] = comp
                            layer.components[i_comp].id_ = id_


    def from_drawio(self, drawio) -> None:
        for tier in drawio.tiers:
            stackml_tier = Tier(name=tier.name, type_=tier.type)
            self.tiers.append(stackml_tier)
            for node in tier.nodes:
                stackml_node = stackml_tier.add_node(id_=node.name, cardinality=1, type_=node.type)
                for layer in reversed(node.layers):
                    if len(layer.components) > 0:
                        stackml_layer = stackml_node.add_layer_element(element='', type_=layer.type, hub=layer.hub)
                        for component in layer.components:
                            stackml_layer.add_component(Component(id_=component.name,
                                                                  type_=component.type.split('Component')[0]))

        for link in drawio.links:
            source = drawio.get_node(link.source)
            sink = drawio.get_node(link.source)
            stackml_link = Link(source=source.name,
                                sink=sink.name, layer=link.Layer)
            self.links.append(stackml_link)

    def sort_tiers(self):
        tiers_name = ['Thing', 'Edge', 'Network', 'Platform', 'Application']
        sorted_tiers = []
        for tier in tiers_name:
            tier_ = self.get_tier(tier)
            if tier_:
                sorted_tiers.append(tier_)

        self.tiers = sorted_tiers
